#!/bin/bash

PID=$$
SCRIPT=$(readlink -f "$0")
DIR=$(cd "$(dirname $SCRIPT)" && pwd)
HOMEDIR=$(getent passwd pi | cut -d: -f6)

data_dir="$DIR/data"
states_path="$data_dir/.states"
log_path="$DIR/log/CAFCA.log"
ra_log="$DIR/log/retroarch/retroarch.log"

game_info=$(grep -i -m1 'ACTIVE_GAME' "$data_dir/.settings" | cut -d\' -f2)
cafca_action=$(grep -w -m1 'ACTION' "$data_dir/.settings" | cut -d\' -f2)
coin_key=$(grep -w -m1 '^COIN_KEY' "$data_dir/.settings" | cut -d\' -f2)
start_key=$(grep -w -m1 '^START_KEY' "$data_dir/.settings" | cut -d\' -f2)

teensy_id='/dev/input/by-id/usb-Teensyduino_Serial_Keyboard_Mouse_Joystick_1008140-if02-event-kbd'
serial_fifo='/tmp/serial.fifo'

exception_list=$(grep -w -m1 '^EXCEPTION_LIST' "$data_dir/.settings" | cut -d\' -f2)

declare -Ag data=(); declare -ag regs=(); declare -ag keys=()

# -----------------------------------------------------------------------------------

HAS_DATA=0
GAME_READY=0
IDENTIFIED=0

ADDR=''
VAL=0
LAST_VAL=

# -----------------------------------------------------------------------------------

echo -e "CAFCA $(date '+%F %T')\n"

sudo cp -r "$data_dir/.states.default" "$data_dir/.states"
[ ! -f "$data_dir/credits.log" ] && touch "$data_dir/credits.log"; sudo chown pi:root "$data_dir/credits.log"

T=$((10#$(date +%s)))

printf "C:0\nS:0\nX:0\nT:$T\n" > "$data_dir/credits.log"
printf "$T\n" > "$data_dir/tmp/start_time"

# -----------------------------------------------------------------------------------

Main() {
  echo "ACTION  $cafca_action"
  STATE GAME_LOADED 1
  getData
  waitForGame
  STATE GAME_READY 1

  if [[ $cafca_action == SCAN ]]; then
    sleep 5
    for state in {1..0}; do evemu-event "$teensy_id" --type EV_KEY --code $coin_key --value $state --sync; sleep 0.2; done; sleep 2
    #pressKey "$coin_key"; sleep 1
    scan_coin; checkMatches
  else
    while [[ $(grep -wE -m1 '^X' "$data_dir/credits.log" | cut -d: -f2) -lt 1 ]]; do :; done
    echo -e "CREDITS 1" > "$serial_fifo"

    if [[ $HAS_DATA -gt 0 ]]; then
      [[ "${data[REGNAM]}" == misc ]] && { checkMiscReg; } || { ADDR=$(dump_addr); IDENTIFIED=1; }
      [[ "${data[GAME]}" == @(${exception_list}) ]] && checkException $ADDR &
    fi

    while [[ $(pgrep -c retroarch) -gt 0 ]]; do

      if [[ $IDENTIFIED -gt 0 ]]; then
        VAL=$(sudo scanmem -p `pidof retroarch` -c"dump $ADDR 1;exit" 2>&1 | awk 'NR==15 {print $2}')
        [[ "$VAL" =~ ^[0-9A-Fa-f]+$ ]] && VAL=$VAL || VAL=$LAST_VAL
      else
        VAL=$(grep -wE -m1 '^X' "$data_dir/credits.log" | cut -d: -f2)
      fi

      if [[ $VAL != $LAST_VAL ]]; then
        [[ "$VAL" =~ ^[0-9A-Fa-f]+$ ]] && CREDITS=$( echo "$((16#$VAL))") || { [[ "$VAL" =~ ^[0-9]+$ ]] && CREDITS="$VAL"; }
        echo -e "CREDITS $CREDITS" > "$serial_fifo"
        echo "$( (( ${#ADDR} )) && echo $ADDR || echo COINS ) : $VAL ( $CREDITS )"
        LAST_VAL=$VAL
      fi
      sleep 1
    done
  fi
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

getData() {
  data[GAME]="${game_info#*'/'}"; data[SYSTEM]="${game_info%'/'*}"
  addr_data=$(grep -w -m1 "^${data[GAME]}" "$data_dir/${data[SYSTEM]}/table" | awk '{$1=$1="";print}')

  if [[ $( echo "${addr_data}" | wc -c ) -gt 1 ]]; then
    for i in OFFSET REGNUM REGNAM VARPOS DATYPE; do (( count++ ));
      val=$(echo "$addr_data" | awk -v pos="$count" '{print $pos}'); data["$i"]="$val";
    done
    HAS_DATA=1
    for n in GAME SYSTEM OFFSET REGNUM REGNAM VARPOS DATYPE; do printf "%s\t%s\n" "$n" "${data[$n]}"; done; echo ""
  else
    echo -e "GAME:   ${data[GAME]}\nSYSTEM: ${data[SYSTEM]}\n - NO DATA FOUND!"
  fi
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

dump_addr() {
  scan_args="option region_scan_level 3; reset; dregions !${data[REGNUM]}; lregions; exit"
  REG_OFFS=$(sudo scanmem -p `pidof retroarch` -c"$scan_args" 2>&1 |& grep --line-buffered -A1 "lregions" | awk 'NR==2' | awk '$1=$1 {print}' | cut -d] -f2 | awk -F" " '{print $1}' | cut -d, -f1);REG_OFFS="0x${REG_OFFS}"
  printf -v ADDR "0x%X\n" $(( REG_OFFS + ${data[OFFSET]} )) &>/dev/null; echo "${ADDR,,}"
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

checkMiscReg() {
  regs+=("${data[REGNUM]}")
  for i in {1..10}; do regs+=("$(( ${data[REGNUM]} - i ))"); regs+=("$(( ${data[REGNUM]} + i ))"); done; #echo -e "regs: ${regs[@]})"

  for reg in "${regs[@]}"; do
    data[REGNUM]="$reg"; ADDR=$(dump_addr)
    VAL=$(sudo scanmem -p `pidof retroarch` -c"dump $ADDR 1;exit" 2>&1 | awk 'NR==15' | awk -F' ' '{print $2}')
    COINS=$(grep -E -m1 '^X' "$data_dir/credits.log" | cut -d: -f2)
    [[ "$VAL" =~ ^[0-9A-Fa-f]+$ ]] && [[ $(echo "$((16#$VAL))") -eq $COINS ]] && IDENTIFIED=1 || IDENTIFIED=0

    echo -e "REGNUM: ${data[REGNUM]}, ADDR: $ADDR, VAL: $VAL\n"
    [[ $IDENTIFIED -gt 0 ]] && break
  done
  [[ $IDENTIFIED -gt 0 ]] && echo "MATCH FOUND! Waiting for START press.." || echo "NO MATCH.."

  while [[ $(grep -E -m1 '^S' "$data_dir/credits.log" | cut -d: -f2) -lt 1 ]]; do :; done

  VAL=$(sudo scanmem -p `pidof retroarch` -c"dump $ADDR 1;exit" 2>&1 | awk 'NR==15' | awk -F' ' '{print $2}')
  COINS=$(grep -E -m1 '^X' "$data_dir/credits.log" | cut -d: -f2)
  [[ "$VAL" =~ ^[0-9A-Fa-f]+$ ]] && [[ $(echo "$((16#$VAL))") -eq $COINS ]] && IDENTIFIED=1 || IDENTIFIED=0

  echo -e "\nREGNUM: ${data[REGNUM]}, ADDR: $ADDR, VAL: $VAL "
  [[ $IDENTIFIED -gt 0 ]] && echo "- IDENTIFIED!" || echo "- NO MATCH.."
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

scan_coin() {
  printf '' > "$DIR/log/expect.log"

  /usr/bin/expect -c '
    log_user 1
    exp_internal 0

    set timeout -1
    set teensy_id '/dev/input/by-id/usb-Teensyduino_Serial_Keyboard_Mouse_Joystick_1008140-if02-event-kbd'
    set coin_key "KEY_C"
    set coin_file '/home/pi/CAFCA/data/credits.log'
    set exp_log '/home/pi/CAFCA/log/expect.log'
    log_file "$exp_log"

    # ------------------------------------------------------------

    set pid [exec pidof retroarch]
    set ready 0
    set credits 1
    set retries 0
    set last_credits 0
    set matches 999
    set matches_last 998
    set match_limit 3

    # ------------------------------------------------------------

    spawn scanmem -p $pid

    proc wait { time } {
      send_user "waiting $time seconds..\n"
      set t_bak $::timeout
      set ::timeout $time
      expect -exact "<%SLEEPTIMER%>" { timeout abort }
      set ::timeout $t_bak
    }

    proc startscan {} {
      expect {
        "Please" {
          sleep 1
          send "option region_scan_level 3\rreset\r"
          sleep 1
          exp_continue
        }
        "*suitable*" {
          set ::ready 1
          sleep 1
        }
      }

      while {$::matches > $::match_limit} {
        get_coin

        if {$::matches < 6} {
          if {$::matches == $::matches_last} {
            set ::retries [expr {$::retries + 1}]
            if {$::retries > 3} { send "exit\r"; break }
          } else {
            send "list\r"
          }
        }
        set ::matches_last $::matches
      }
      sleep 1
      #send "list\r"

      expect "* 0*" {
        sleep 1
        send "exit\r"
      }
    }

    proc get_coin {} {
      set ::credits [exec grep -m1 '^C' "$::coin_file" | cut -d: -f2 ]
      #set ::credits [exec printf "%02X" $::credits ]
      sleep 1

      if {$::credits != $::last_credits} {
        set ::last_credits $::credits
        sleep 1
        send "$::credits\r"
        sleep 1

        expect {
          "*currently*" {
            set ::matches [lindex [split [lindex [split $expect_out(buffer) "\n"] end-1 ] " " ] end-1 ]
            sleep 1

            if {$::matches == "0"} {
              set ::matches 999
              send "option region_scan_level 3\rreset\r$::credits\r"
            } elseif {$::matches == "other"} {
              set ::matches 1
            } else {
              insert_coin
            }
          }
        }
      }
    }

    proc insert_coin {} {
      exec evemu-event $::teensy_id --type EV_KEY --code $::coin_key --value 1 --sync
      exec evemu-event $::teensy_id --type EV_KEY --code $::coin_key --value 0 --sync
    }

    startscan
    close
    expect eof
    wait
  '
  sudo sed -i 's/\r//g' "$DIR/log/expect.log"
  echo ""
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

checkMatches() {
  local LINE=; local ADDR=; local VAL=; local LAST_VAL=; local REGNAM=; local DUMP=''; local IDENTIFIED=0
  local COINS=$(grep -wE -m1 '^X' "$DIR/data/credits.log" | cut -d: -f2)
  local match_data=$(tac "$DIR/log/expect.log" | grep -Pn -m1 '^\[.*?[[:digit:]]\]' | awk -F '\[|\]' '{gsub(/\ /, "", $2);print $1$2}')
  local LN=$(echo "$match_data" | cut -d: -f1); local VL=$(echo "$match_data" | cut -d: -f2)
  declare -a MATCHES=()

  echo -e "$(grep ACTIVE_GAME $DIR/data/.settings | cut -d\' -f2):\n" > "$DIR/data/tmp/matches.log" ; echo -e "COINS: $COINS\n"

  for (( i=LN; i<=$(( LN + VL )); i++)); do
    LINE=$(tac "$DIR/log/expect.log" | awk -F '\[.*?[[:digit:]].*?|\]\ ' -v ln="$i" 'NR==ln {print $2}' | awk '{gsub(/\ \+/,",");gsub(/  +/," "); sub(/^ /,"");$1="0x"$1;$3="0x"$3" "$2;$2="";print}' | sed 's/ //; s/, / /g; s/ \]/\]/g' | sed -E ':a; s/(\[[^]\ ]*)[[:blank:]] */\1\,/; ta; s/\[([^]]*)\]/\1/g')
    ADDR=$(echo "$LINE" | awk '{print $1}'); VAL=$(sudo scanmem -p `pidof retroarch` -c"dump $ADDR 1;exit" 2>&1 | awk 'NR==15 {print $2}')
    [[ $((16#$VAL)) -eq $COINS ]] && { MATCHES+=("$LINE $VAL")
    echo -e "MATCH $(( i - LN )) = $ADDR : $VAL"; }; echo ""
  done

  #pressKey "$start_key"; sleep 2
  for state in {1..0}; do evemu-event "$teensy_id" --type EV_KEY --code $start_key --value $state --sync; sleep 0.2; done; sleep 2
  COINS=$(grep -wE -m1 '^X' "$DIR/data/credits.log" | cut -d: -f2)
  echo -e "COINS: $COINS\n"

  for MATCH in "${MATCHES[@]}"; do
    ADDR=$(echo "$MATCH" | awk '{print $1}'); REGNAM=$(echo "$MATCH" | awk '{print $4}')
    LAST_VAL=$(echo "$MATCH" | grep -o '[^ ]*$'); IDENTIFIED=1

    echo -e "${MATCH}:\n"

    for i in {1..3}; do
      VAL=$(sudo scanmem -p `pidof retroarch` -c"dump $ADDR 1;exit" 2>&1 | awk 'NR==15 {print $2}')
      [[ $((16#$VAL)) -ne $(( LAST_VAL - 1 )) ]] && IDENTIFIED=0
      sleep 0.5
      echo -e "VAL: $VAL ( $((16#$VAL)) )"
    done

    if [[ $IDENTIFIED -gt 0 ]]; then
      DUMP=$(echo "${MATCH}" | cut -d' ' -f2- | awk '{$NF=""; print $0}')
      echo "${DUMP}" >> "$DIR/data/tmp/matches.log"; echo -e " - IDENTIFIED!\n"
      [[ $REGNAM == code ]] && break
    fi
  done

  if [[ $IDENTIFIED -gt 0 ]]; then
    if [[ $(grep -w -m1 "^${data[GAME]}" "$data_dir/${data[SYSTEM]}/table" | awk '{print NF}') -le 1 ]]; then
      sudo cp "$data_dir/${data[SYSTEM]}/table" "$data_dir/${data[SYSTEM]}/.bak/table.bak$(date +%y%m%d%H%M%S)"
      sudo sed -i "/${data[GAME]}/ s/\(.\)$/ ${DUMP}/" "$data_dir/${data[SYSTEM]}/table"
    fi
    tac  "$DIR/log/matches.log" | "$DIR/log/matches.log" &>/dev/null
    echo "$(grep -m1 ^${data[GAME]} $data_dir/${data[SYSTEM]}/table) ($(date +%Y-%m-%d_%H-%M-%S))" | sudo tee -a "$DIR/log/matches.log" &>/dev/null
    tac  "$DIR/log/matches.log" | "$DIR/log/matches.log" &>/dev/null
  fi

 echo "DONE"

  # RETURN WITHOUT WRITING
  #return
  #if [[ $IDENTIFIED -gt 0 ]]; then
  #  if [[ $(grep -w -m1 "^${data[GAME]}" "$data_dir/${data[SYSTEM]}/table" | awk '{print NF}') -le 1 ]]; then
  #    sudo cp "$data_dir/${data[SYSTEM]}/table" "$data_dir/${data[SYSTEM]}/table.bak$(date +%y%m%d%H%M%S)"
  #    ROW=$(echo "$(grep -w -m1 ^${data[GAME]} $data_dir/${data[SYSTEM]}/table)$(echo $LINE | cut -d' ' -f2-)")
  #    #sudo sed -i "s/${data[GAME]}/${ROW}/" "$data_dir/${data[SYSTEM]}/table"
  #    sudo sed -i "/s/${data[GAME]}/ s/\(.\)$/ ${str}/" "$data_dir/${data[SYSTEM]}/table"
  #  fi
  #fi
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

pressKey() {
  [[ ${#keys[@]} -lt 1 ]] && keys=($(sudo timeout 1 evtest "$teensy_id" |& awk '{$1=$1;print}' | grep '^Event' | awk -F'\(|\)' '{print $2}' | grep '^KEY'))

  if [[ $# -gt 0 ]]; then
    local KEY="$@"
    [[ "${KEYS[@]}" =~ "$KEY" ]] && for state in {1..0}; do evemu-event "$teensy_id" --type EV_KEY --code "$KEY" --value "$state" --sync; sleep 0.2; done
  fi
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

waitForGame() {
  local words=('hiscore.dat' 'SET_GEOMETRY')
  local T_LOAD=$(grep -w -m1 "^${data[GAME]}" "$data_dir/${data[SYSTEM]}/timers" | grep -o '[^ ]*$')

  while [[ $GAME_READY -lt 1 ]]; do GAME_READY=$(cat "$ra_log" | grep -Ec "$(for w in $(echo ${words[@]} | tr ' ' ' | '); do echo *$w*; done)"); done
  while [[ $((10#$(date +%s)-T)) -lt $T_LOAD ]]; do :; done

  echo -e "\nGAME_READY!\n"
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

STATE() {
  [ $# -lt 2 ] && return
  local KEY="${1^^}"; local VAL="${2}"; local SER=${3:-1}
  [[ $(printf "${VAL}" | grep -Ec "[[:alpha:]]") -gt 0 ]] && VAL="'${VAL}'"

  if [[ $(grep -ic "$KEY" "$data_dir/.states") -gt 0 ]]; then
    sudo sed -i "s/^$KEY.*/$KEY=$VAL/g" "$data_dir/.states"
    [[ $SER -gt 0 ]] && echo -e "$KEY $VAL" > "$serial_fifo"
  fi
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

exitFunc() {
  echo -e "GAME_STOPPED 1" >> "$serial_fifo"
  STATE 'GAME_READY' '0'; STATE 'GAME_LOADED' '0'

  if [[ "${data[SYSTEM]}" == mame-libretro ]]; then
    local clearRam=$(grep -m1 "^${data[GAME]}" "$data_dir/${data[SYSTEM]}/clearram" | awk '{print $2}')
    local nvRamPath="$HOMEDIR/RetroPie/roms/${data[SYSTEM]}/mame2003/nvram/${data[GAME]}.nv"
    [[ $clearRam -gt 0 ]] && [ -f "$nvRamPath" ] && sudo rm "$nvRamPath"
  fi

  JOBS=($(jobs -p)); [[ ${#JOBS[@]} -gt 0 ]] && sudo kill -9 "${JOBS[@]}" > /dev/null 2>&1
  exit
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

checkException() {
  local param1="${1}"
  local param2="${2}"

  case "${data[GAME]}" in
    'blktiger')
      local coin_state=1
      local acc_addr=$(printf "0x%x\n" $(( ADDR - 0xe40 )))
      local acc_val=

      while [[ $(pgrep -c retroarch) -gt 0 ]]; do
        acc_val=$(sudo scanmem -p `pidof retroarch` -c"dump $acc_addr 1;exit" 2>&1 | awk 'NR==15' | awk '{print $2}')
        [[ $acc_val == 00 ]] && [[ $coin_state -eq 0 ]] && { echo "COIN_POWER 1"; echo -e "COIN_POWER 1" > /tmp/serial.fifo; coin_state=1; } || { \
        [[ $acc_val == 20 ]] && [[ $coin_state -eq 1 ]] && { echo "COIN_POWER 0"; echo -e "COIN_POWER 0" > /tmp/serial.fifo; coin_state=0; }; }
      done
      ;;
    *)
      ;;
  esac
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

trap 'exitFunc' EXIT SIGINT INT

Main

exit 0











# ___________________________________________________________________________________
#
# ˅ IGNORE ˅
